import logging

# External libraries
import ruamel.yaml

from .common import CWL_VERSIONS, DEF_VERSION, CWL_SHEBANG, Namespaces, Parameter
from .requirements import *
from .utils import literal, literal_presenter, Serializable

logging.basicConfig(level=logging.INFO)
_LOGGER = logging.getLogger(__name__)

#  Function(s)  ------------------------------

#  Class(es)  ------------------------------


class CommandLineTool(Serializable):
    '''
    Contain all informations to describe a CWL command line tool.
    '''

    __CLASS__ = 'CommandLineTool'

    def __init__(self, tool_id=None, base_command=None, label=None, doc=None,
                 cwl_version=None, stdin=None, stderr=None, stdout=None, path=None):
        '''
        :param tool_id: unique identifier for this tool
        :type tool_id: STRING
        :param base_command: command line for the tool
        :type base_command: STRING | list[STRING]
        :param label: label of this tool
        :type label: STRING
        :param doc: documentation for the tool, usually longer than the label
        :type doc: STRING
        :param cwl_version: version of the CWL tool
        :type cwl_version: STRING
        :param stdin: path to a file whose contents must be piped into stdin
        :type stdin: STRING
        :param stderr: capture stderr into the given file
        :type stderr: STRING
        :param stdout: capture stdout into the given file
        :type stdout: STRING

        inputs (:class:`cwlgen.CommandInputParameter` objects),
        outputs (:class:`cwlgen.CommandOutputParameter` objects),
        arguments (:class:`cwlgen.CommandLineBinding` objects),
        and requirements (:class:`cwlgen.Requirement` objects)
        are stored in lists which are initialized empty.
        '''
        if cwl_version not in CWL_VERSIONS:
            _LOGGER.warning("CWL version {} is not recognized as a valid version.".format(cwl_version))
            _LOGGER.warning("CWL version is set up to {}.".format(DEF_VERSION))
            cwl_version = DEF_VERSION
        self.cwlVersion = cwl_version
        self.id = tool_id
        self.label = label
        self.requirements = []  # List of Several object inhereting from [Requirement]
        self.hints = []
        self.inputs = []  # List of [CommandInputParameter] objects
        self.outputs = []  # List of [CommandOutputParameter] objects
        self.baseCommand = base_command
        self.arguments = []  # List of [CommandLineBinding] objects
        self.doc = doc
        self.stdin = stdin
        self.stderr = stderr
        self.stdout = stdout
        self.successCodes = []
        self.temporaryFailCodes = []
        self.permanentFailCodes = []
        self._path = path
        self.namespaces = Namespaces()
        self.metadata = {}
        self.ignore_attributes = ["namespaces"]

    def get_dict(self):

        d = super(CommandLineTool, self).get_dict()

        d['class'] = self.__CLASS__

        if self.metadata:
            for key, value in self.metadata.__dict__.items():
                d["s:" + key] = value
            # - Add Namespaces
            d[self.namespaces.name] = {}
            for k, v in self.namespaces.__dict__.items():
                if '$' not in v:
                    d[self.namespaces.name][k] = v

        if self.requirements:
            d['requirements'] = {r.get_class(): r.get_dict() for r in self.requirements}
        return d

    def export(self, outfile=None):
        """
        Export the tool in CWL either on STDOUT or in outfile.
        """
        # First add representer (see .utils.py) for multiline writting
        ruamel.yaml.add_representer(literal, literal_presenter)

        cwl_tool = self.get_dict()

        # Write CWL file in YAML
        if outfile is None:
            six.print_(CWL_SHEBANG, "\n", sep='')
            six.print_(ruamel.yaml.dump(cwl_tool))
        else:
            out_write = open(outfile, 'w')
            out_write.write(CWL_SHEBANG + '\n\n')
            out_write.write(ruamel.yaml.dump(cwl_tool))
            out_write.close()


class CommandInputParameter(Parameter):
    '''
    An input parameter for a :class:`cwlgen.CommandLineTool`.
    '''

    def __init__(self, param_id, label=None, secondary_files=None, param_format=None,
                 streamable=None, doc=None, input_binding=None, default=None, param_type=None):
        '''
        :param param_id: unique identifier for this parameter
        :type param_id: STRING
        :param label: short, human-readable label
        :type label: STRING
        :param secondary_files: If type is a file, describes files that must be
                                included alongside the primary file(s)
        :type secondary_files: STRING
        :param param_format: If type is a file, uri to ontology of the format or exact format.
        :type param_format: STRING
        :param streamable: If type is a file, true indicates that the file is read or written
                           sequentially without seeking
        :type streamable: BOOLEAN
        :param doc: documentation
        :type doc: STRING
        :param input_binding: describes how to handle the input
        :type input_binding: :class:`cwlgen.CommandLineBinding` object
        :param default: default value
        :type default: STRING
        :param param_type: type of data assigned to the parameter corresponding to CWLType
        :type param_type: STRING
        '''
        Parameter.__init__(self, param_id=param_id, label=label,
                           secondary_files=secondary_files, param_format=param_format,
                           streamable=streamable, doc=doc, param_type=param_type)
        self.inputBinding = input_binding
        self.default = default


class CommandOutputParameter(Parameter):
    '''
    An output parameter for a :class:`cwlgen.CommandLineTool`.
    '''

    def __init__(self, param_id, label=None, secondary_files=None, param_format=None,
                 streamable=None, doc=None, output_binding=None, param_type=None):
        '''
        :param param_id: unique identifier for this parameter
        :type param_id: STRING
        :param label: short, human-readable label
        :type label: STRING
        :param secondary_files: If type is a file, describes files that must be
                                included alongside the primary file(s)
        :type secondary_files: STRING
        :param param_format: If type is a file, uri to ontology of the format or exact format
        :type param_format: STRING
        :param streamable: If type is a file, true indicates that the file is read or written
                           sequentially without seeking
        :type streamable: BOOLEAN
        :param doc: documentation
        :type doc: STRING
        :param output_binding: describes how to handle the output
        :type output_binding: :class:`cwlgen.CommandOutputBinding` object
        :param param_type: type of data assigned to the parameter corresponding to CWLType
        :type param_type: STRING
        '''
        Parameter.__init__(self, param_id, label, secondary_files, param_format, streamable,
                           doc, param_type)
        self.outputBinding = output_binding


class CommandOutputBinding(Serializable):
    '''
    Describes how to generate an output parameter based on the files produced.
    '''

    def __init__(self, glob=None, load_contents=None, output_eval=None):
        '''
        :param glob: Find corresponding file(s)
        :type glob: STRING
        :param load_contents: For each file matched, read up to the 1st 64 KiB of text and
                              place it in the contents field
        :type load_contents: BOOLEAN
        :param output_eval: Evaluate an expression to generate the output value
        :type output_eval: STRING
        '''
        self.glob = glob
        self.loadContents = load_contents
        self.outputEval = output_eval



import unittest
import cwlgen


class TestAddRequirements(unittest.TestCase):

    def test_inlinejs(self):
        w = cwlgen.Workflow()
        req = cwlgen.InlineJavascriptReq(["expression"])
        w.requirements.append(req)
        d = w.get_dict()
        self.assertIn("requirements", d)
        dr = d["requirements"]
        self.assertIn(req.get_class(), dr)
        drr = dr[req.get_class()]
        self.assertIn("expressionLib", drr)
        self.assertEqual(drr["expressionLib"], ["expression"])

    def test_commandtool(self):
        c = cwlgen.CommandLineTool("reqs")
        docker = cwlgen.DockerRequirement(docker_pull="ubuntu:latest")
        c.requirements.append(docker)

        d = c.get_dict()
        self.assertIn("requirements", d)
        self.assertIn(docker.get_class(), d["requirements"])


class TestInlineJavascriptReq(unittest.TestCase):

    def setUp(self):
        self.js_req = cwlgen.InlineJavascriptReq(expression_lib=['expression'])
        self.js_req_nolib = cwlgen.InlineJavascriptReq()

    def test_init(self):
        self.assertEqual(self.js_req.get_class(), 'InlineJavascriptRequirement')
        self.assertEqual(self.js_req.expressionLib, ['expression'])
        self.assertEqual(self.js_req_nolib.get_class(), 'InlineJavascriptRequirement')
        self.assertIsNone(self.js_req_nolib.expressionLib)

    def test_export(self):
        tool = self.js_req.get_dict()
        self.assertEqual(tool, {'expressionLib': ['expression']})

    def test_export_without_lib(self):
        tool = self.js_req_nolib.get_dict()
        self.assertEqual(tool, {})


class TestDockerRequirement(unittest.TestCase):

    def setUp(self):
        self.dock_req = cwlgen.DockerRequirement(
            docker_pull='pull',
            docker_load='load',
            docker_file='file',
            docker_import='import',
            docker_image_id='id',
            docker_output_dir='dir'
        )

    def test_init(self):
        self.assertEqual(self.dock_req.get_class(), 'DockerRequirement')
        self.assertEqual(self.dock_req.dockerPull, 'pull')
        self.assertEqual(self.dock_req.dockerLoad, 'load')
        self.assertEqual(self.dock_req.dockerFile, 'file')
        self.assertEqual(self.dock_req.dockerImport, 'import')
        self.assertEqual(self.dock_req.dockerImageId, 'id')
        self.assertEqual(self.dock_req.dockerOutputDir, 'dir')

    def test_export(self):
        d = self.dock_req.get_dict()
        comparison = {
            'dockerFile': 'file',
            'dockerImageId': 'id',
            'dockerImport': 'import',
            'dockerLoad': 'load',
            'dockerOutputDir': 'dir',
            'dockerPull': 'pull'
        }
        self.assertEqual(d, comparison)

    def test_missing_export(self):
        dr = cwlgen.DockerRequirement()
        self.assertEqual(dr.get_dict(), {})


class TestSubworkflowFeatureRequirement(unittest.TestCase):

    def setUp(self):
        self.req = cwlgen.SubworkflowFeatureRequirement()

    def test_init(self):
        self.assertEqual(self.req.get_class(), 'SubworkflowFeatureRequirement')

    def test_add(self):
        tool = self.req.get_dict()
        self.assertEqual(tool, {})


class TestSoftwareRequirement(unittest.TestCase):

    def setUp(self):
        self.r = cwlgen.SoftwareRequirement()
        self.r.packages.append(cwlgen.SoftwareRequirement.SoftwarePackage("package", "version", None))

    def test_properties(self):
        self.assertEqual(self.r.get_class(), "SoftwareRequirement")
        self.assertEqual(len(self.r.packages), 1)
        p = self.r.packages[0]
        self.assertEqual(p.package, "package")
        self.assertEqual(p.version, "version")
        self.assertIsNone(p.specs)

    def test_serialization(self):
        d = self.r.get_dict()
        self.assertIn("packages", d)
        pks = d["packages"]
        self.assertEqual(len(pks), 1)
        p = pks[0]
        self.assertEqual(p.get("package"), "package")
        self.assertEqual(p.get("version"), "version")
        self.assertNotIn("specs", p)


class TestInitialWorkDirRequirement(unittest.TestCase):

    def setUp(self):
        self.i1 = cwlgen.InitialWorkDirRequirement("listing1")
        self.i2 = cwlgen.InitialWorkDirRequirement([
            cwlgen.InitialWorkDirRequirement.Dirent("entry", "entryname")
        ])

    def test_properties_i1(self):
        self.assertEqual(self.i1.get_class(), "InitialWorkDirRequirement")
        self.assertEqual(self.i1.listing, "listing1")

    def test_properties_i2(self):
        self.assertEqual(len(self.i2.listing), 1)
        l = self.i2.listing[0]
        self.assertEqual(l.entry, "entry")
        self.assertEqual(l.entryname, "entryname")
        self.assertIsNone(l.writable)

    def test_serialization_i1(self):
        d = self.i1.get_dict()
        self.assertIn("listing", d)
        self.assertIsInstance(d["listing"], str)
        self.assertEqual(d["listing"], "listing1")

    def test_serialization_i2(self):
        d = self.i2.get_dict()
        self.assertIn("listing", d)
        listing = d.get("listing")
        self.assertIsInstance(listing, list)
        self.assertEqual(len(listing), 1)
        l = listing[0]
        self.assertIn("entry", l)
        self.assertIn("entryname", l)
        self.assertNotIn("writable", l)

        self.assertEqual(l["entry"], "entry")
        self.assertEqual(l["entryname"], "entryname")


class TestScatterFeatureRequirement(unittest.TestCase):

    def setUp(self):
        self.req = cwlgen.ScatterFeatureRequirement()

    def test_init(self):
        self.assertEqual(self.req.get_class(), 'ScatterFeatureRequirement')

    def test_add(self):
        tool = self.req.get_dict()
        self.assertDictEqual(tool, {})


class TestStepInputExpressionRequirement(unittest.TestCase):

    def setUp(self):
        self.req = cwlgen.StepInputExpressionRequirement()

    def test_init(self):
        self.assertEqual(self.req.get_class(), 'StepInputExpressionRequirement')

    def test_add(self):
        tool = self.req.get_dict()
        self.assertDictEqual(tool, {})


class TestEnvVarRequirement(unittest.TestCase):

    def setUp(self):
        self.e = cwlgen.EnvVarRequirement([
            cwlgen.EnvVarRequirement.EnvironmentDef("env_name", "env_value")
        ])

    def test_properties(self):
        self.assertEqual(self.e.get_class(), "EnvVarRequirement")
        self.assertIsInstance(self.e.envDef, list)
        self.assertEqual(len(self.e.envDef), 1)
        e = self.e.envDef[0]
        self.assertEqual(e.envName, "env_name")
        self.assertEqual(e.envValue, "env_value")

    def test_serialization(self):
        d = self.e.get_dict()
        self.assertIn("envDef", d)
        ed = d["envDef"]
        self.assertIsInstance(ed, list)
        self.assertEqual(len(ed), 1)
        e = ed[0]
        self.assertEqual(e.get("envName"), "env_name")
        self.assertEqual(e.get("envValue"), "env_value")


class TestShellCommandRequirement(unittest.TestCase):

    def setUp(self):
        self.req = cwlgen.ShellCommandRequirement()

    def test_init(self):
        self.assertEqual(self.req.get_class(), 'ShellCommandRequirement')

    def test_add(self):
        tool = self.req.get_dict()
        self.assertDictEqual(tool, {})


class TestResourceRequirement(unittest.TestCase):

    def setUp(self):
        self.req = cwlgen.ResourceRequirement(
            cores_min=1, cores_max=3,
            ram_min=2, ram_max=4,
            tmpdir_min="tmp/min", tmpdir_max="tmp/max",
            outdir_min="out/min", outdir_max="out/max"
        )
        self.empty = cwlgen.ResourceRequirement()

    def test_init(self):
        self.assertEqual(self.req.get_class(), 'ResourceRequirement')

        self.assertEqual(1, self.req.coresMin)
        self.assertEqual(3, self.req.coresMax)

        self.assertEqual(2, self.req.ramMin)
        self.assertEqual(4, self.req.ramMax)

        self.assertEqual("tmp/min", self.req.tmpdirMin)
        self.assertEqual("tmp/max", self.req.tmpdirMax)
        self.assertEqual("out/min", self.req.outdirMin)
        self.assertEqual("out/max", self.req.outdirMax)

    def test_empty(self):
        self.assertDictEqual(self.empty.get_dict(), {})

    def test_add(self):
        r = self.req.get_dict()
        self.assertEqual(r.get("coresMin"), 1)
        self.assertEqual(r.get("coresMax"), 3)

        self.assertEqual(r.get("ramMin"), 2)
        self.assertEqual(r.get("ramMax"), 4)

        self.assertEqual(r.get("tmpdirMin"), "tmp/min")
        self.assertEqual(r.get("tmpdirMax"), "tmp/max")
        self.assertEqual(r.get("outdirMin"), "out/min")
        self.assertEqual(r.get("outdirMax"), "out/max")




'''
Library to handle the manipulation and generation of CWL tool
'''

#  Import  ------------------------------

# General libraries
import logging

# External libraries
import ruamel.yaml
import six
from .version import __version__

from .utils import literal, literal_presenter

from .import_cwl import parse_cwl

logging.basicConfig(level=logging.INFO)
_LOGGER = logging.getLogger(__name__)

# imports for __init__

from .common import *
from .commandlinetool import *
from .workflow import *
from .commandlinebinding import CommandLineBinding
from .requirements import *

#  Function(s)  ------------------------------

#  Class(es)  ------------------------------




import unittest


class TestExport(unittest.TestCase):

    def test_workflow_export(self):
        import cwlgen
        w = cwlgen.Workflow("identifier")
        w.export()

    def test_commandlinetool_export(self):
        import cwlgen

        c = cwlgen.CommandLineTool("identifier", "echo")
        c.export()



from .utils import Serializable


class CommandLineBinding(Serializable):
    """
    The binding behavior when building the command line depends on the data type of the value.
    If there is a mismatch between the type described by the input schema and the effective value,
    such as resulting from an expression evaluation, an implementation must use the data type of the effective value.

    Documentation: https://www.commonwl.org/v1.0/CommandLineTool.html#CommandLineBinding
    """

    def __init__(self, load_contents=None, position=None, prefix=None, separate=None,
                 item_separator=None, value_from=None, shell_quote=None):
        """
        :param load_contents: Read up to the fist 64 KiB of text from the file and
                              place it in the "contents" field of the file object
        :type load_contents: BOOLEAN
        :param position: The sorting key
        :type position: INT
        :param prefix: Command line prefix to add before the value
        :type prefix: STRING
        :param separate:
        :type separate: BOOLEAN
        :param item_separator: Join the array elements into a single string separated by this item
        :type item_separator: STRING
        :param value_from: Use this as the value
        :type value_from: STRING
        :param shell_quote: Value is quoted on the command line
        :type shell_quote: BOOLEAN
        """
        self.loadContents = load_contents
        self.position = position
        self.prefix = prefix
        self.separate = separate
        self.itemSeparator = item_separator
        self.valueFrom = value_from
        self.shellQuote = shell_quote



from setuptools import setup

__version__ = "v0.0.2"

setup(name="illusional.cwlgen",
        version=__version__,
        description='Fork of python-cwlgen | Generation of CWL programmatically. Avaible types: CommandLineTool, Workflow and Requirements',
        long_description=open("./README.md").read(),
        author='Michael Franklin | Kenzo-Hugo Hillion and Herve Menager',
        author_email='michael.franklin@petermac.org',
        license='MIT',
        keywords = ['cwl'],
        install_requires=['six', 'ruamel.yaml==0.13.13'],
        packages=["cwlgen"],
        classifiers=[
            'Development Status :: 4 - Beta',
            'Topic :: Scientific/Engineering',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'Environment :: Console',
            ],
        )



import logging

from .utils import Serializable

logging.basicConfig(level=logging.INFO)
_LOGGER = logging.getLogger(__name__)

#  Constant(s)  ------------------------------

CWL_SHEBANG = "#!/usr/bin/env cwl-runner"
CWL_VERSIONS = ['draft-2', 'draft-3.dev1', 'draft-3.dev2', 'draft-3.dev3',
                'draft-3.dev4', 'draft-3.dev5', 'draft-3', 'draft-4.dev1',
                'draft-4.dev2', 'draft-4.dev3', 'v1.0.dev4', 'v1.0']
DEF_VERSION = 'v1.0'


#  Function(s)  ------------------------------

def parse_type(param_type, requires_type=False):
    """
    Parses the parameter type as one of the required types:
    :param requires_type:
    :: https://www.commonwl.org/v1.0/CommandLineTool.html#CommandInputParameter

    :param param_type: a CWL type that is _validated_
    :type param_type: CWLType | CommandInputRecordSchema | CommandInputEnumSchema | CommandInputArraySchema | string |
       array<CWLType | CommandInputRecordSchema | CommandInputEnumSchema | CommandInputArraySchema | string>
    :return: CWLType | CommandInputRecordSchema | CommandInputEnumSchema | CommandInputArraySchema | string |
       array<CWLType | CommandInputRecordSchema | CommandInputEnumSchema | CommandInputArraySchema | string>
    """

    if requires_type is False and param_type is None:
        return None

    if isinstance(param_type, str) and len(param_type) > 0:
        # Must be CWLType
        optional = param_type[-1] == "?"
        if optional:
            _LOGGER.debug("Detected {param_type} to be optional".format(param_type=param_type))
            cwltype = param_type[:-1]
        else:
            cwltype = param_type
        # cwltype = param_type[:-1] if optional else param_type

        # check for arrays
        if len(cwltype) > 2 and cwltype[-2:] == "[]":
            array_type = CommandInputArraySchema(items=cwltype[:-2])
            # How to make arrays optional input: https://www.biostars.org/p/233562/#234089
            return [CwlTypes.DEF_TYPE, array_type] if optional else array_type

        if cwltype not in CwlTypes.TYPES:
            _LOGGER.warning("The type '{param_type}' is not a valid CWLType, expected one of: {types}"
                            .format(param_type=param_type, types=", ".join(str(x) for x in CwlTypes.TYPES)))
            _LOGGER.warning("type is set to {}.".format(CwlTypes.DEF_TYPE))
            return CwlTypes.DEF_TYPE
        return param_type

    elif isinstance(param_type, list):
        return [parse_type(p) for p in param_type]

    elif isinstance(param_type, CommandInputArraySchema) \
            or isinstance(param_type, CommandInputRecordSchema) \
            or isinstance(param_type, CommandInputEnumSchema):
        return param_type  # validate if required

    if requires_type is True:
        raise Exception("'parse_type' was required but failed to parse '{ptype}', exiting")

    _LOGGER.warning("Unable to detect type of param '{param_type}'".format(param_type=param_type))
    return CwlTypes.DEF_TYPE


def get_type_dict(param_type):
    """
    Convets
    :param param_type:
    :type param_type: CWLType | InputRecordSchema | InputEnumSchema | InputArraySchema | string |
                array<CWLType | InputRecordSchema | InputEnumSchema | InputArraySchema | string>
    :return: str | dict
    """
    if isinstance(param_type, str):
        return param_type
    elif isinstance(param_type, list):
        return [get_type_dict(p) for p in param_type]
    elif isinstance(param_type, dict):
        return param_type
    elif getattr(param_type, 'get_dict', None) and callable(getattr(param_type, 'get_dict', None)):
        return param_type.get_dict()
    else:
        raise Exception("Could not convert '{param_type}' to dictionary as it was unrecognised"
                        .format(param_type=type(param_type)))


#  Class(es)  ------------------------------

class CwlTypes:
    DEF_TYPE = "null"

    NULL = "null"
    BOOLEAN = "boolean"
    INT = "int"
    LONG = "long"
    FLOAT = "float"
    DOUBLE = "double"
    STRING = "string"
    FILE = "File"
    DIRECTORY = "Directory"
    STDOUT = "stdout"
    ARRAY = "array"

    NON_NULL_TYPES = [BOOLEAN, INT, LONG, FLOAT, DOUBLE, STRING, FILE, DIRECTORY, STDOUT]
    TYPES = [NULL, None, BOOLEAN, INT, LONG, FLOAT, DOUBLE, STRING, FILE, DIRECTORY, STDOUT]


# functions

class Namespaces(object):
    """
    Define different namespace for the description.
    """

    def __init__(self):
        """
        """
        self.name = "$namespaces"
        self.s = "http://schema.org/"


class Metadata(object):
    """
    Represent metadata described by http://schema.org.
    """

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


class Parameter(Serializable):
    '''
    Based class for parameters (common field of Input and Output) for CommandLineTool and Workflow
    '''

    def __init__(self, param_id, label=None, secondary_files=None, param_format=None,
                 streamable=None, doc=None, param_type=None, requires_type=False):
        '''
        :param param_id: unique identifier for this parameter
        :type param_id: STRING
        :param label: short, human-readable label
        :type label: STRING
        :param secondary_files: If type is a file, describes files that must be
                                included alongside the primary file(s)
        :type secondary_files: STRING
        :param param_format: If type is a file, uri to ontology of the format or exact format
        :type param_format: STRING
        :param streamable: If type is a file, true indicates that the file is read or written
                           sequentially without seeking
        :type streamable: BOOLEAN
        :param doc: documentation
        :type doc: STRING
        :param param_type: type of data assigned to the parameter
        :type param_type: STRING corresponding to CWLType
        '''

        self.id = param_id
        self.label = label
        self.secondaryFiles = secondary_files
        self.format = param_format
        self.streamable = streamable
        self.doc = doc
        self.type = parse_type(param_type, requires_type)


class CommandInputArraySchema(Serializable):
    '''
    Based on the parameter set out in the CWL spec:
    https://www.commonwl.org/v1.0/CommandLineTool.html#CommandInputArraySchema
    '''

    def __init__(self, items=None, label=None, input_binding=None):
        '''
        :param items: Defines the type of the array elements.
        :type: `CWLType | CommandInputRecordSchema | CommandInputEnumSchema | CommandInputArraySchema | string | array<CWLType | CommandInputRecordSchema | CommandInputEnumSchema | CommandInputArraySchema | string>`
        :param label: A short, human-readable label of this object.
        :type label: STRING
        :param input_binding:
        :type input_binding: CommandLineBinding
        '''
        self.type = CwlTypes.ARRAY
        self.items = parse_type(items, requires_type=True)
        self.label = label
        self.inputBinding = input_binding


class CommandInputRecordSchema(Serializable):
    """
    Documentation: https://www.commonwl.org/v1.0/Workflow.html#CommandInputRecordSchema
    """

    def __init__(self, label=None, name=None):
        """
        :param fields: Defines the fields of the record.
        :type fields: array<InputRecordField>
        :param label: A short, human-readable label of this object.
        :param name: NF (Name of the InputRecord)
        """
        self.fields = []
        self.label = label
        self.name = name
        self.type = "record"

    class CommandInputRecordField(Serializable):
        """
        Documentation: https://www.commonwl.org/v1.0/Workflow.html#CommandInputRecordField
        """

        def __init__(self, name, input_type, doc=None, input_binding=None, label=None):
            """
            :param name:
            :param input_type:
            :type input_type: CWLType | InputRecordSchema | InputEnumSchema | InputArraySchema | string |
                        array<CWLType | InputRecordSchema | InputEnumSchema | InputArraySchema | string>
            :param doc: A documentation string for this field
            :param input_binding:
            :type input_binding: CommandLineBinding
            :param label:
            """
            self.name = name
            self.type = parse_type(input_type, requires_type=True)
            self.doc = doc
            self.inputBinding = input_binding
            self.label = label

        def get_dict(self):
            d = super(type(self), self).get_dict()
            d["type"] = get_type_dict(self.type)
            return d


class CommandInputEnumSchema(Serializable):
    """
    Documentation: https://www.commonwl.org/v1.0/Workflow.html#CommandInputEnumSchema
    """

    def __init__(self, symbols, label=None, name=None, input_binding=None):
        """
        :param symbols: Defines the set of valid symbols.
        :type symbols: List[str]
        :param label: A short, human-readable label of this object.
        :type label: str
        :param name:
        :type name: str
        :param input_binding:
        :type input_binding: CommandLineBinding
        """
        self.type = "enum"
        self.symbols = symbols
        self.label = label
        self.name = name
        self.inputBinding = input_binding
